import { World, AnyEntity, Loop, Debugger } from "@rbxts/matter"
import Plasma from "@rbxts/plasma"
import { Players, ReplicatedStorage, RunService, UserInputService } from "@rbxts/services"

import { Host } from "ReplicatedStorage/hosts"
import { tags } from "./boundTags"
import { Model } from "./components"
import { start as startReplication, stop as stopReplication } from "./replication"
import { ClientState, ServerState } from "./state"
import { start as startSystems, stop as stopSystems } from "./systems"
import { start as startTags, stop as stopTags } from "./tags"

const MAX_DISPLAY_ORDER = 2147483647

function authorize(player: Player): boolean {
    return RunService.IsStudio() || game.CreatorId === player.UserId
}

let connections:
    | {
        [index: string]: RBXScriptConnection
    }
    | undefined

export function start<S extends object>(host: Host, state: S): [World, S] {
    if (connections) throw "ECS already running."

    const world = new World()
    const debug = new Debugger(Plasma)
    debug.authorize = authorize

    debug.findInstanceFromEntity = (id: AnyEntity): Instance | undefined => {
        if (!world.contains(id)) return
        const model = world.get(id, Model)
        return model?.model
    }

    const loop = new Loop(world, state, debug.getWidgets())
    startSystems(host, loop, debug)
    debug.autoInitialize(loop)

    connections = loop.begin({
        default: RunService.Heartbeat,
        Stepped: RunService.Stepped
    })

    if (host === Host.All || host === Host.Server) {
        const ServerState = state as ServerState

        startTags(world, tags, ServerState)
    }

    if (host === Host.All || host === Host.Client) {
        const ClientState = state as ClientState

        startReplication(world, ClientState)

        const serverDebugger = ReplicatedStorage.FindFirstChild("MatterDebugger")
        if (serverDebugger && serverDebugger.IsA("ScreenGui")) {
            serverDebugger.DisplayOrder = MAX_DISPLAY_ORDER
        }

        const clientDebugger = Players.LocalPlayer.FindFirstChild("MatterDebugger")
        if (clientDebugger && clientDebugger.IsA("ScreenGui")) {
            clientDebugger.DisplayOrder = MAX_DISPLAY_ORDER
        }

        UserInputService.InputBegan.Connect((input) => {
            if (input.KeyCode === Enum.KeyCode.F4 && authorize(Players.LocalPlayer)) {
                debug.toggle()
                ClientState.debugEnabled = debug.enabled
            }
        })
    }

    return [world, state]
}

/**
 * Stops the ECS.
 */
export function stop(): void {
	if (!connections) return
	for (const [_, connection] of pairs(connections)) {
		connection.Disconnect()
	}
	connections = undefined
	stopTags()
	stopReplication()
	stopSystems()
	ReplicatedStorage.FindFirstChild("MatterDebuggerRemote")?.Destroy()
}